#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <signal.h>
#include <sys/wait.h>
#ifdef ENABLE_NLS
#  include <libintl.h>
#  undef _
#  define _(String) dgettext (PACKAGE, String)
#  ifdef gettext_noop
#    define N_(String) gettext_noop (String)
#  else
#    define N_(String) (String)
#  endif
#else
#  define textdomain(String) (String)
#  define gettext(String) (String)
#  define dgettext(Domain,Message) (Message)
#  define dcgettext(Domain,Message,Type) (Message)
#  define bindtextdomain(Domain,Directory) (Domain)
#  define _(String) (String)
#  define N_(String) (String)
#endif
#include "baytech.p.h"
#include <sys/types.h>
#include <sys/wait.h>

typedef struct _node_info Node_Info;

struct _node_info
{
   int on;   
   int emp;
   int va1k;
};

GtkWidget *base = NULL;

void
on_on1_activate                        (GtkMenuItem     *menuitem,
					gpointer         user_data);
void
on_off1_activate                       (GtkMenuItem     *menuitem,
					gpointer         user_data);
GtkWidget *create_base(void);
GtkWidget *create_pixmap(GtkWidget *w, char *file);
void add_pixmap_directory(char *dir);

void
flim_client_message(Msg *msg)
{
   if (!strcmp(msg->msg, "nexxus_add"))
     {
	Nexxus *nx;
	char nx_name[4096], nx_address[4096], nx_login[4096], nx_password[4096];
	
	nx = nexxus_new();
	nx_login[0] = 0;
	nx_password[0] = 0;
	sscanf(msg->dat, "%[^:]:%[^:]:%[^:]:%[^:]", nx_name, nx_address, nx_login, nx_password);
	nx->name = g_strdup(nx_name);
	nx->address = g_strdup(nx_address);
	nx->login = g_strdup(nx_login);
	nx->password = g_strdup(nx_password);
	nexxus_connect(nx);
     }
   else if (!strcmp(msg->msg, "nexxus_del"))
     {
	Nexxus *nx;
	
	nx = nexxus_find(msg->dat);
	if (nx) 
	  {
	     nexxus_free(nx);
	  }
     }
   else if (!strcmp(msg->msg, "nexxus_rename"))
     {
	Nexxus *nx;
	char nx_name[4096], nx_name2[4096];
	
	sscanf(msg->dat, "%[^:]:%[^:]", nx_name, nx_name2);
	nx = nexxus_find(nx_name);
	if (nx)
	  {
	     if (nx->name) g_free(nx->name);
	     nx->name = g_strdup(nx_name2);
	  }
     }
   else if (!strcmp(msg->msg, "nexxus_address"))
     {
	Nexxus *nx;
	char nx_name[4096], nx_addr[4096];
	
	sscanf(msg->dat, "%[^:]:%[^:]", nx_name, nx_addr);
	nx = nexxus_find(nx_name);
	if (nx)
	  {
	     if (nx->address) g_free(nx->address);
	     nx->address = g_strdup(nx_addr);
	     if (nx->connection) flim_comms_disconnect_nexxus(nx->connection);
	     nx->connection = NULL;
	     nexxus_connect(nx);
	  }
     }
   else if (!strcmp(msg->msg, "node_add"))
     {
	Nexxus *nx;
	Node *nd;
	char nx_name[4096], nd_name[4096];
	
	sscanf(msg->dat, "%[^:]:%[^:]", nx_name, nd_name);
	nx = nexxus_find(nx_name);
	if (nx)
	  {
	     Node_Info *ni;
	     
	     nd = node_new(nx);
	     nd->name = g_strdup(nd_name);
	     ni = g_malloc(sizeof(Node_Info));
	     ni->on = 0;
	     ni->emp = 0;
	     ni->va1k = 0;
	     nd->data = ni;
	  }
     }
   else if (!strcmp(msg->msg, "node_del"))
     {
	Node *nd;
	
	nd = node_find(msg->dat);
	if (nd)
	  {
	     Node_Info *ni;
	     
	     ni = nd->data;
	     g_free(ni);
	     node_free(nd);
	  }
     }
   else if (!strcmp(msg->msg, "node_rename"))
     {
	Node *nd;
	char nx_name[4096], nd_name[4096], nd_name2[4096], buf[4096];
	
	sscanf(msg->dat, "%[^:]:%[^:]:%[^:]", nx_name, nd_name, nd_name2);
	g_snprintf(buf, sizeof(buf), "%s:%s", nx_name, nd_name);
	nd = node_find(buf);
	if (nd)
	  {
             if (nd->name) g_free(nd->name);
	     nd->name = g_strdup(nd_name2);	     
	  }
     }
   else if (!strcmp(msg->msg, "nexxus_select"))
     {
	Nexxus *nx;

	nx = nexxus_find(msg->dat);
	if (nx) 
	   {
	      nx->selected = 1;
	   }
     }
   else if (!strcmp(msg->msg, "node_select"))
     {
	Node *nd;

	nd = node_find(msg->dat);
	if (nd) 
	  {
	     nd->selected = 1;
	  }
     }
   else if (!strcmp(msg->msg, "nexxus_unselect"))
     {
	Nexxus *nx;

	nx = nexxus_find(msg->dat);
	if (nx) 
	  {
	     nx->selected = 0;
	  }
     }
   else if (!strcmp(msg->msg, "node_unselect"))
     {
	Node *nd;

	nd = node_find(msg->dat);
	if (nd) 
	  {
	     nd->selected = 0;
	  }
     }
}

void
on_on1_activate                        (GtkMenuItem     *menuitem,
					gpointer         user_data)
{
   GList *l;
   
   for (l = nexxus_list; l; l = l->next)
     {
	Nexxus *nx;
	GList *ll;
	
	nx = l->data;
	for (ll = nx->nodes; ll; ll = ll->next)
	  {
	     Node *nd;
	     
	     nd = ll->data;
	     if (nd->selected)
	       {
		  char buf[4096];
		  
		  g_snprintf(buf, sizeof(buf), "BAYTECH:PORT_ON:%s:%i", nd->name, (int)user_data);
		  FLIM_COMMS_BEGIN(nx->connection, buf);
		  while (FLIM_COMMS_REPLY)
		    {
		       if (!strcmp("JOB_COMPLETED", FLIM_COMMS_REPLY[2]))
			  break;
		       if (!strcmp("JOB_ERROR", FLIM_COMMS_REPLY[2]))
			  break;
		       FILM_COMMS_GET_REPLY;
		    }
		  FILM_COMMS_REPLY_END;
	       }
	  }
     }
   menuitem = NULL;
   user_data = NULL;
}


void
on_off1_activate                       (GtkMenuItem     *menuitem,
					gpointer         user_data)
{
   GList *l;
   
   for (l = nexxus_list; l; l = l->next)
     {
	Nexxus *nx;
	GList *ll;
	
	nx = l->data;
	for (ll = nx->nodes; ll; ll = ll->next)
	  {
	     Node *nd;
	     
	     nd = ll->data;
	     if (nd->selected)
	       {
	     Node *nd;
	     
	     nd = ll->data;
	     if (nd->selected)
	       {
		  char buf[4096];
		  
		  g_snprintf(buf, sizeof(buf), "BAYTECH:PORT_OFF:%s:%i", nd->name, (int)user_data);
		  FLIM_COMMS_BEGIN(nx->connection, buf);
		  while (FLIM_COMMS_REPLY)
		    {
		       if (!strcmp("JOB_COMPLETED", FLIM_COMMS_REPLY[2]))
			  break;
		       if (!strcmp("JOB_ERROR", FLIM_COMMS_REPLY[2]))
			  break;
		       FILM_COMMS_GET_REPLY;
		    }
		  FILM_COMMS_REPLY_END;
	       }
	       }
	  }
     }
   menuitem = NULL;
   user_data = NULL;
}

GtkWidget *
create_base(void)
{
   GtkWidget *base;
   GtkWidget *menubar1;
   GtkWidget *b1;
   GtkWidget *b1_menu;
   GtkAccelGroup *b1_menu_accels;
   GtkWidget *port_1;
   GtkWidget *port_1_menu;
   GtkAccelGroup *port_1_menu_accels;
   GtkWidget *on1;
   GtkWidget *off1;
   GtkWidget *pixmap1;
   
   base = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   gtk_object_set_data (GTK_OBJECT (base), "base", base);
   gtk_container_set_border_width (GTK_CONTAINER (base), 0);
   gtk_window_set_title (GTK_WINDOW (base), "Flim");
   gtk_window_set_wmclass (GTK_WINDOW (base), "Sercon", "FLIM");
   
   menubar1 = gtk_menu_bar_new ();
   gtk_widget_ref (menubar1);
   gtk_object_set_data_full (GTK_OBJECT (base), "menubar1", menubar1,
			     (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (menubar1);
   gtk_container_add (GTK_CONTAINER (base), menubar1);
   
   b1 = gtk_menu_item_new();
   gtk_widget_ref (b1);
   gtk_widget_show (b1);
   gtk_container_add (GTK_CONTAINER (menubar1), b1);
   
   pixmap1 = create_pixmap(base, "baytech.xpm");
   gtk_widget_ref(pixmap1);
   gtk_object_set_data_full (GTK_OBJECT (base), "pixmap1", base,
			     (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (pixmap1);
   gtk_container_add (GTK_CONTAINER (b1), pixmap1);
   
     {
	GtkTooltips *tt;
	
	tt = gtk_tooltips_new();
	gtk_tooltips_set_delay(GTK_TOOLTIPS(tt), 300);
	gtk_tooltips_enable(GTK_TOOLTIPS(tt));
	gtk_tooltips_set_tip(GTK_TOOLTIPS(tt), b1, 
			     "Power on/off baytech power ports"
			     , "");
     }
   
   b1_menu = gtk_menu_new ();
   gtk_widget_ref (b1_menu);
   gtk_menu_item_set_submenu (GTK_MENU_ITEM (b1), b1_menu);
   b1_menu_accels = gtk_menu_ensure_uline_accel_group (GTK_MENU (b1_menu));
#define PORT(name, number) \
port_1 = gtk_menu_item_new_with_label (name); \
gtk_widget_show (port_1); \
gtk_container_add (GTK_CONTAINER (b1_menu), port_1); \
port_1_menu = gtk_menu_new (); \
gtk_menu_item_set_submenu (GTK_MENU_ITEM (port_1), port_1_menu); \
port_1_menu_accels = gtk_menu_ensure_uline_accel_group (GTK_MENU (port_1_menu)); \
on1 = gtk_menu_item_new_with_label ("On"); \
gtk_widget_show (on1); \
gtk_container_add (GTK_CONTAINER (port_1_menu), on1); \
off1 = gtk_menu_item_new_with_label ("Off"); \
gtk_widget_show (off1); \
gtk_container_add (GTK_CONTAINER (port_1_menu), off1); \
gtk_signal_connect (GTK_OBJECT (on1), "activate", \
GTK_SIGNAL_FUNC (on_on1_activate), \
(gpointer)number); \
gtk_signal_connect (GTK_OBJECT (off1), "activate", \
GTK_SIGNAL_FUNC (on_off1_activate), \
(gpointer)number);
   
/* FIXME: need to wait for nexxus to get ipc to query number of ports */
   PORT("Port 1", 0);
   PORT("Port 2", 1);
   PORT("Port 3", 2);
   PORT("Port 4", 3);
   PORT("Port 5", 4);
   PORT("Port 6", 5);
   PORT("Port 7", 6);
   PORT("Port 8", 7);
   PORT("Port 9", 8);
   PORT("Port 10", 9);
   PORT("Port 11", 10);
   PORT("Port 12", 11);
   PORT("Port 13", 12);
   PORT("Port 14", 13);
   PORT("Port 15", 14);
   PORT("Port 16", 15);

   
   gtk_signal_connect (GTK_OBJECT (base), "destroy",
		       GTK_SIGNAL_FUNC (exit),
		       NULL);   
   return base;
}

int
main(int argc, char **argv)
{
   Msg msg;
   
   flim_comms_do_gui(0);
   flim_client_init(argc, argv);
#ifdef ENABLE_NLS
   bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
   textdomain (PACKAGE);
#endif
   
   gtk_set_locale ();
   gtk_init (&argc, &argv);
   
   add_pixmap_directory (PACKAGE_DATA_DIR "/pixmaps");
   add_pixmap_directory (PACKAGE_SOURCE_DIR "/pixmaps");
   
   base = create_base();
   gtk_widget_realize(base);

   g_snprintf(msg.msg, sizeof(msg.msg), "is_toolbar");
   flim_client_send(&msg);   
   
   g_snprintf(msg.msg, sizeof(msg.msg), "window_id");
   g_snprintf(msg.dat, sizeof(msg.dat), "%i", (int)GDK_WINDOW_XWINDOW(base->window));
   flim_client_send(&msg);

   for (;;) { XEvent xev; if (XCheckTypedWindowEvent(GDK_WINDOW_XDISPLAY(base->window), GDK_WINDOW_XWINDOW(base->window), ReparentNotify, &xev)) break; usleep(100000); }
   gtk_widget_show(base);

   g_snprintf(msg.msg, sizeof(msg.msg), "ready");
   flim_client_send(&msg);
   
   gtk_main();
   return 0;
}
