#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <signal.h>
#include <sys/wait.h>
#include "client.h"
#include "comms.h"

void flim_client_message(Msg *msg);

static int plugin_id = -1;
static int flim_id = -1;
static int msg_id = -1;
static int accept_messages = 0;

typedef struct _msgbuf Flim_MsgBuf;

struct _msgbuf
{
   long mtype;
   char mtext[5120];
};

GList *nexxus_list = NULL;

Nexxus *
nexxus_new(void)
{
   Nexxus *nx;
   
   nx = g_malloc(sizeof(Nexxus));
   nx->name = NULL;
   nx->address = NULL;
   nx->selected = 0;
   nx->data = NULL;
   nx->nodes = NULL;
   nx->connection = NULL;
   nx->login = NULL;
   nx->password = NULL;
   nexxus_list = g_list_append(nexxus_list, nx);
   return nx;
}

void
nexxus_connect(Nexxus *nx)
{
   nx->connection = flim_comms_connect_nexxus(nx->address, nx->login, nx->password);
}

Node *
node_new(Nexxus *nx)
{
   Node *nd;
   
   nd = g_malloc(sizeof(Node));
   nd->name = NULL;
   nd->selected = 0;
   nd->data = NULL;
   nd->nexxus = nx;
   nx->nodes = g_list_append(nx->nodes, nd);
   return nd;
}

void
nexxus_free(Nexxus *nx)
{
   GList *l;
   
   if (nx->connection) flim_comms_disconnect_nexxus(nx->connection);
   if (nx->name) g_free(nx->name);
   if (nx->address) g_free(nx->address);
   if (nx->login) g_free(nx->login);
   if (nx->password) g_free(nx->password);
   l = nx->nodes;
   while(l)
     {
	Node *nd;
	
	nd = l->data;
	l = l->next;
	node_free(nd);
     }
   nexxus_list = g_list_remove(nexxus_list, nx);
   g_free(nx);
}

void
node_free(Node *nd)
{
   nd->nexxus->nodes = g_list_remove(nd->nexxus->nodes, nd);
   if (nd->name) g_free(nd->name);
   g_free(nd);
}

Nexxus *
nexxus_find(char *name)
{
   GList *l;
   
   for (l = nexxus_list; l; l = l->next)
     {
	Nexxus *nx;

	nx = l->data;
	if (!strcmp(nx->name, name)) return nx;
     }
   return NULL;
}

Node *
node_find(char *name)
{
   GList *l;
   char nx_name[4096], nd_name[4096];
   
   sscanf(name, "%[^:]:%[^:]", nx_name, nd_name);
   for (l = nexxus_list; l; l = l->next)
     {
	Nexxus *nx;
	GList *ll;
	
	nx = l->data;
	if (!strcmp(nx_name, nx->name))
	  {
	     for (ll = nx->nodes; ll; ll = ll->next)
	       {
		  Node *nd;
		  
		  nd = ll->data;
		  if (!strcmp(nd_name, nd->name)) return nd;
	       }
	  }
     }
   return NULL;
}

static void
int_handle_child(int num)
{
   pid_t               pid;
   int                 status;
   
   num = 0;
   while ((pid = waitpid(-1, &status, WNOHANG)) > 0);
   return;
}

static void
int_handle_exit(int num)
{
   num = 0;
   exit(0);
}

static gint
flim_client_msg_timeout(gpointer data)
{
   if (accept_messages > 0) return 1;
   flim_client_handle_messages();
   return 1;
   data = NULL;
}

static void
flim_client_cleanup(void)
{
   msgctl(msg_id, IPC_RMID, NULL);
}

void
flim_client_init(int argc, char **argv)
{
   Msg msg;
   struct sigaction    sa;
   
   sa.sa_handler = int_handle_child; sa.sa_flags = SA_RESTART; sigemptyset(&sa.sa_mask);
   sigaction(SIGCHLD, &sa, (struct sigaction *)0);
   sa.sa_handler = int_handle_exit; sa.sa_flags = SA_RESTART; sigemptyset(&sa.sa_mask);
   sigaction(SIGINT, &sa, (struct sigaction *)0);
   sa.sa_handler = int_handle_exit; sa.sa_flags = SA_RESTART; sigemptyset(&sa.sa_mask);
   sigaction(SIGQUIT, &sa, (struct sigaction *)0);
   sa.sa_handler = int_handle_exit; sa.sa_flags = SA_RESTART; sigemptyset(&sa.sa_mask);
   sigaction(SIGTERM, &sa, (struct sigaction *)0);
   if (argc > 1) flim_id = atoi(argv[1]);
   if (argc > 2) plugin_id = atoi(argv[2]);
   atexit(flim_client_cleanup);
   msg_id = msgget(ftok(getenv("HOME"), plugin_id), IPC_CREAT | 0666);
   g_snprintf(msg.msg, sizeof(msg.msg), "return_ipc");
   g_snprintf(msg.dat, sizeof(msg.dat), "%i", msg_id);
   flim_client_send(&msg);
   gtk_timeout_add(50, flim_client_msg_timeout, NULL);
}

void
flim_client_send(Msg *msg)
{
   Flim_MsgBuf msgp;
   
   if (flim_id < 0) return;
   msgp.mtype = 1;
   msg->id = plugin_id;
   memcpy(&(msgp.mtext[0]), msg, sizeof(Msg));
   msgsnd(flim_id, &msgp, sizeof(Msg), 0);
}

void
flim_client_handle_messages(void)
{
   Flim_MsgBuf msgp;
   
   msgp.mtype = 1;
   msgp.mtext[0] = 0;
   while (msgrcv(msg_id, &msgp, 1024, 0, IPC_NOWAIT) >= 0)
     {
	Msg *msg;
	
	msg = (Msg *)&(msgp.mtext[0]);
	flim_client_message(msg);
     }
}

void
flim_client_block_messages(void)
{
   accept_messages++;
}

void
flim_client_unblock_messages(void)
{
   accept_messages--;
}
