/*
 * This program is original work by San Mehat <nettwerk@valinux.com> and
 * Zac Sprackett <zac@sprackett.com>.
 * This program is distributed under the GNU Lesser General Public License
 * (LGPL).
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 */

#ifndef _VACMCLIENT_API_H
#define _VACMCLIENT_API_H

#ifdef __cplusplus
extern "C" {
#endif

/*
 * API function return codes
 */
#define API_RETURN_OK									0
#define API_RETURN_BAD_HANDLE					-1
#define API_RETURN_NEXXUS_DISCONNECT	-2
#define API_RETURN_LOOKUP_FAILED			-3
#define API_RETURN_BAD_ADDRESS				-4
#define API_RETURN_MISC_ERROR					-5
#define API_RETURN_ALREADY_CONNECTED	-6
#define API_RETURN_TIMED_OUT					-7
#define API_RETURN_AUTH_FAILED				-8

/*
 * Prototypes for api functions
 */

int api_nexxus_connect(char	*addr,
											 char	*username,
											 char	*password,
											 void	**new_handle);
int api_nexxus_disconnect(void *handle);
int api_nexxus_send_ipc(void				*handle,
												char				*pkt,
												 __uint32_t	len);
int api_nexxus_recv_ipc(void				*handle,
												char				**buffer,
												__uint32_t	*len);
int api_nexxus_wait_for_data(void				*handle,
														 char				**buffer, 
														 __uint32_t	*len,
														 int				timeout);
int api_nexxus_wait_short_for_data(void				*handle,
														 char				**buffer, 
														 __uint32_t	*len,
														 int				timeout);

int api_nexxus_return_fd(void *handle,
												 int	*fd);
int api_nexxus_return_ip(void						*handle,
												struct in_addr	*ip_addr);
int api_nexxus_return_handle(void **handle,
														 char	*addr);
void api_nexxus_perror(char *msg);
char *api_nexxus_get_error(void);
int api_nexxus_ping(struct in_addr *ip_addr, struct timeval *to);

	/*
	 *	Shuts off error messages for the various calls.
	 */
extern int	api_nexxus_suppress_error_messages;

#ifdef __cplusplus
} /* __cplusplus */
#endif
#endif
