/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 */

#include <stdio.h>
#include <sys/socket.h>
#include <signal.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <stdlib.h>
#include <sys/time.h>
#include <dlfcn.h>
#include <errno.h>
#include <nexxus.h>
#include <prototypes.h>
#include <pthread.h>
#include <config.h>

typedef struct client_filedesc_lock
	{
	int							fd;
	pthread_mutex_t	lock;

	struct	client_filedesc_lock	*next;
	struct	client_filedesc_lock	*prev;
	} CLIENT_FD_LOCK;

static CLIENT_FD_LOCK	*locklist_head = NULL;
static CLIENT_FD_LOCK	*locklist_tail = NULL;

extern int	nexxus_debug;
static CLIENT_FD_LOCK	*lookup_lock(int fd);

int client_fd_lock_create(int fd)
{
  CLIENT_FD_LOCK  *new;
	char						tmp_str[255];

	sprintf(tmp_str,"client fd %d lock",fd);
if (nexxus_debug)
	nexxus_log_printf(NULL, "Creating FD lock for %d\n",fd);

  if (!(new = (CLIENT_FD_LOCK *) nettmcs_alloc (sizeof(CLIENT_FD_LOCK),
                                              	tmp_str)))
    {
    nexxus_log_printf(NULL,"Unable to allocate memory for client FD lock\n");
    return(-1);
    }
  memset(new, 0, sizeof(CLIENT_FD_LOCK));

  if (!locklist_tail)
    locklist_tail = locklist_head = new;
  else
    {
    locklist_tail->next = new;
    new->prev = locklist_tail;
    }
  locklist_tail = new;
	new->fd = fd;
	pthread_mutex_init(&new->lock, NULL);

  return(0);
}

void client_fd_lock_destroy(int fd)
{
	CLIENT_FD_LOCK	*lock;

	if (!(lock = lookup_lock(fd)))
		{
		nexxus_log_printf(NULL,"No FD lock for fd %d!\n",fd);
		return;
		}
	if (nexxus_debug)
		nexxus_log_printf(NULL, "Destroying FD lock for %d\n",fd);

  if (lock == locklist_head)
    {
    locklist_head = locklist_head->next;
    if (locklist_head)
      locklist_head->prev = NULL;
    else
      locklist_tail = NULL;
    }
  else if (lock == locklist_tail)
    {
    locklist_tail = locklist_tail->prev;
    if (locklist_tail)
      locklist_tail->next = NULL;
    else
      locklist_head = NULL;
    }
  else
    {
    lock->next->prev = lock->prev;
    lock->prev->next = lock->next;
    }
	pthread_mutex_unlock(&lock->lock); // Just to be sure 
  nettmcs_free(lock);
}

static CLIENT_FD_LOCK	*lookup_lock(int fd)
{
	CLIENT_FD_LOCK	*lock;

	lock = locklist_head;
	while(lock)
		{
		if (lock->fd == fd)
			return(lock);
		lock = lock->next;
		}
	return(NULL);
}

void nexxus_do_lock(int fd)
{
	CLIENT_FD_LOCK	*lock;
	
	if (!(lock = lookup_lock(fd)))
		{
		nexxus_log_printf(NULL,"Unable to lock fd %d as no FD lock exists!\n",fd);
		return;
		}
	if (nexxus_debug)
		nexxus_log_printf(NULL, "Locking FD %d\n",fd);
	pthread_mutex_lock(&lock->lock);
	if (nexxus_debug)
		nexxus_log_printf(NULL, "Locked FD %d\n",fd);
}

void nexxus_do_unlock(int fd)
{
	CLIENT_FD_LOCK	*lock;
	
	if (!(lock = lookup_lock(fd)))
		{
		nexxus_log_printf(NULL,"Unable to unlock fd %d as no FD lock exists!\n",fd);
		return;
		}
	if (nexxus_debug)
		nexxus_log_printf(NULL, "Unlocking FD %d\n",fd);
	pthread_mutex_unlock(&lock->lock);
	if (nexxus_debug)
		nexxus_log_printf(NULL, "Unlocked FD %d\n",fd);
}
