
/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <pthread.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#ifndef __sgi
#include <linux/types.h>
#include <linux/serial.h>
#endif
#include <libloose.h>
#include <emp_module_harness.h>
#include <emp_slave_context.h>
#include <prototypes.h>
#include <emp_slave_context.h>
#include <pthread.h>
#include <emp.h>
#include <emp_sdr.h>
#include <emp_sel.h>
#include <emp_sensor_events.h>

static int send_driver_ipc(CONTEXT *ctx, int fd, DRIVER_IPC *msg);
static int process_active(CONTEXT *ctx, unsigned char *frame);
static int	driver_detect_protocol(CONTEXT *, int, int, int);
static int wait_for_byte(CONTEXT *, unsigned char *);
static void driver_thread_cleanup(void *);
static int driver_rearm(CONTEXT *ctx);
static int fake_modem(CONTEXT *ctx);
static int emp_validate_cksum(CONTEXT *ctx,
                              unsigned char *frame,
                              int framelen,
                              int quiet);

/*
#define EMPHW_DEBUG
#define FRU_DEBUG
#define DETECT_DEBUG
#define SEL_DEBUG
#define SDR_DEBUG
#define ENGAGE_DEBUG
#define SENSOR_DEBUG
#define ALL_DEBUG
#define SEL_DEBUG
#define REARM_DEBUG
#define TRANS_DEBUG	
#define NEGO_DEBUG
#define THREAD_DEBUG
#define DRIVER_DEBUG
*/


#ifdef ALL_DEBUG
  #define DRIVER_DEBUG
  #define THREAD_DEBUG
	#define EMPHW_DEBUG
#endif

#ifdef THREAD_DEBUG
	#define EMPHW_DEBUG
  #define SEL_DEBUG
  #define NEGO_DEBUG
	#define ENGAGE_DEBUG
  #define SENSOR_DEBUG
#endif

#ifdef NEGO_DEBUG
  #define SDR_DEBUG
	#define FRU_DEBUG
	#define SEL_DEBUG
#endif

#ifdef DRIVER_DEBUG
  #define DETECT_DEBUG
  #define TRANS_DEBUG
  #define REARM_DEBUG
#endif

#define STATE_SEND_DB			1
#define STATE_RECV_DA			2
#define STATE_RECV_DP			3
#define STATE_SEND_REPLY	4
void *driver_thread_start(void *data)
{
	CONTEXT *ctx = (CONTEXT *) 			data;
	struct _pthread_cleanup_buffer 	_buff;
	struct timeval									to;
	fd_set													read_fds;
	int															max = 0, rc, full_pkt = 0;
	int															unsolicited_recv_pos = 0;
	unsigned char										unsolicited_pkt[EMP_MAX_MSG_LEN *2];
	char														log_msg[255];
	DRIVER_IPC											msg;
	time_t													current;


#ifdef THREAD_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Thread %s driver starting (pid %d)",
					ctx->node_id,
					getpid());
	send_msg(ctx,ctx->thread_ipc_write, log_msg);
#endif

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	ctx->refcount++;
	_pthread_cleanup_push(&_buff, driver_thread_cleanup, (void *) ctx);
	ctx->last_active_timestamp = time(NULL);
	ctx->next_rearm = time(NULL)+30;

	while(1)
		{
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec =0;
		max = 0;

		if (ctx->emp_sock != -1)
			{
			FD_SET(ctx->emp_sock, &read_fds);
			if (ctx->emp_sock > max)
				max = ctx->emp_sock;
			}
		FD_SET(ctx->driver_pread, &read_fds);
		
		if (ctx->driver_pread > max)
			max = ctx->driver_pread;

		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Driver %s select() failed (%s)",
							ctx->node_id,
							strerror(errno));
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			pthread_testcancel();
			continue;
			}
		if (!rc)
			{
			if (!ctx->session_active)
				goto do_driver_checks;
//				continue;
			current = time(NULL);
			if ((current >= ctx->next_rearm) && (ctx->emp_sock != -1))
				{
				int rrc = -EAGAIN;

				while(rrc == -EAGAIN)
					{
#ifdef REARM_DEBUG
					printf("Driver %s Rearming [",ctx->node_id);
#endif
					rrc = driver_rearm(ctx);
#ifdef REARM_DEBUG
					printf("%d]\n",rrc);
#endif
					}
				ctx->next_rearm = time(NULL) + 30;
				}
			pthread_testcancel();
			}
		else if (FD_ISSET(ctx->driver_pread, &read_fds))
			{
			if (read(ctx->driver_pread, (void *) &msg, sizeof(msg)) < 0)
				{
				sprintf(log_msg,
								"0:0:LOG:Driver %s ipc read failed (%s)\n",
								ctx->node_id,
								strerror(errno));
				send_msg(ctx,ctx->thread_ipc_write, log_msg);
				pthread_testcancel();
				continue;
				}
			if (msg.request == DIPCREQ_TRANS)
				{
				DRIVER_IPC		reply;
				int						len, sendpos, recvpos, state;
				int						byte_retry_cnt, frame_retry_cnt;
				unsigned char	raw_recv[EMP_MAX_MSG_LEN*2];
				unsigned char	norm_recv[EMP_MAX_MSG_LEN*2];
				int						full_interrupted_pkt, dont_wait_for_reply;
				int						response_len, recv_abort, full_duplex_ack;
				int						duplex_retry = 0;
retry:
				response_len = recv_abort = full_duplex_ack = 0;
				full_interrupted_pkt = dont_wait_for_reply = 0;
				byte_retry_cnt = frame_retry_cnt = 0;
				len = sendpos = recvpos = 0;
				state = STATE_SEND_DB;
				memset(raw_recv, 0, sizeof(raw_recv));
				memset(norm_recv, 0, sizeof(norm_recv));
				memset(&reply, 0, sizeof(reply));
				reply.request = DIPCREQ_RESP;

				if ((!ctx->session_active) && (msg.cmd != CMD_SUBMIT_PASSWORD))
					{
					reply.rc = ENOTCONN;
					send_driver_ipc(ctx, ctx->driver_pwrite, &reply);
					continue;
					}
				if (ctx->console_mode)
					{
					reply.rc = EUNATCH;
					send_driver_ipc(ctx, ctx->driver_pwrite, &reply);
					continue;
					}
				/* Transmit frame and wait for response */
				len = msg.pkt_len;
				while(1)
					{
					if (state == STATE_SEND_DB)
						{
#ifdef TRANS_DEBUG
						printf("[%s] Sending 0x%.2x\n",ctx->node_id,msg.pkt_buffer[sendpos]);
#endif
						if (write(ctx->emp_sock, &msg.pkt_buffer[sendpos], 1) < 0)
							{
							reply.rc = errno;
							state = STATE_SEND_REPLY;
							continue;
							}
						state = STATE_RECV_DA;
						continue;
						}
					else if (state == STATE_RECV_DA)
						{
						unsigned char	UC_byte;
						int						rc;
				
						if ((rc = wait_for_byte(ctx, &UC_byte)) < 0)
							{
#ifdef TRANS_DEBUG
							printf("[%s] Timed out waiting for ack\n",ctx->node_id);
#endif
							if (rc == -ETIMEDOUT)
								{
								byte_retry_cnt ++;
								if (byte_retry_cnt == 2)
									{
									frame_retry_cnt ++;
									if (frame_retry_cnt == 2)
										{
#ifdef TRANS_DEBUG
										printf("[%s] Cancelling transaction\n",ctx->node_id);
#endif
										reply.rc = ETIMEDOUT;
										state = STATE_SEND_REPLY;
										continue;
										}
#ifdef TRANS_DEBUG
									printf("[%s] Resending frame\n",ctx->node_id);
#endif
									/* Retry the whole frame */
									byte_retry_cnt = 0;
									sendpos = 0;
									state = STATE_SEND_DB;
									continue;
									}
#ifdef TRANS_DEBUG
								printf("[%s] Resending data byte\n",ctx->node_id);
#endif
								state = STATE_SEND_DB;
								continue;
								}
							reply.rc = rc;
							state = STATE_SEND_REPLY;
							continue;
							}
#ifdef TRANS_DEBUG
						printf("[%s] Got ack byte 0x%.2x\n",ctx->node_id,UC_byte);
#endif
						if (UC_byte == (unsigned char) EMP_HANDSHAKE_CHAR)
							{
							sendpos++;
							if (sendpos == len)
								{
#ifdef TRANS_DEBUG
								printf("[%s] SENT\n",ctx->node_id);
#endif
								break;
								}
							state = STATE_SEND_DB;
							}
						else
							{
							if (UC_byte == (unsigned char) EMP_START_CHAR)
								full_interrupted_pkt = 1;
							else if (UC_byte == (unsigned char) EMP_STOP_CHAR)
								continue;
							else
								full_interrupted_pkt = 0;
							if (recvpos == (int) (sizeof(raw_recv)-1))
								{
								printf("BUFFER OVERFLOW!!!(3)\n");
								abort();
								}
							raw_recv[recvpos++] = UC_byte;
							state = STATE_RECV_DP;
							}
						continue;
						}
					else if (state == STATE_RECV_DP)
						{
						unsigned char	UC_byte;
						int						rc, l;
		
						if ((rc = wait_for_byte(ctx, &UC_byte)) < 0)
							{
							reply.rc = rc;
							state = STATE_SEND_REPLY;
							continue;
							}
#ifdef TRANS_DEBUG
						printf("[%s] Got int 0x%.2x\n",ctx->node_id,UC_byte);
#endif
						if (UC_byte == (unsigned char) 0xa8)
							{
#ifdef TRANS_DEBUG
							printf("[%s] Got full duplex ack\n",ctx->node_id);
#endif
							full_duplex_ack = 1;
							continue;
							}
						if (recvpos == (int) (sizeof(raw_recv)-1))
							{
							/* Resynchronize */
#ifdef DETECT_DEBUG
							printf("Driver %s calling detect_protocol as part of resync while DP\n",
											ctx->node_id);
#endif
							if (!(rc = driver_detect_protocol(ctx, 5, 0, 1)))
								{
								reply.request = DIPCREQ_NEG;
								state = STATE_SEND_REPLY;
								continue;
								}
							else if (rc == -EUNATCH)
								{
								ctx->console_mode = 1;
								ctx->enable_driver_checks = 0;
								reply.rc = EUNATCH;
								state = STATE_SEND_REPLY;
								continue;
								}
							reply.rc = EIO;
							state = STATE_SEND_REPLY;
							continue;
							}
						raw_recv[recvpos++] = UC_byte;
						if (UC_byte == (unsigned char) EMP_STOP_CHAR)
							{

							if (full_interrupted_pkt)
								{
								l = normalize_data(&raw_recv[0], &norm_recv[0]);
								if (!emp_validate_cksum(ctx,norm_recv, l, 0))
									{
#ifdef TRANS_DEBUG
									int bleh;
									printf("Interrupted pkt BAD (norm) [");
									for (bleh=0; norm_recv[bleh] != (unsigned char) 0xa5; bleh++)
										printf("0x%.2x ",norm_recv[bleh]);
									printf("0x%.2x]\n",norm_recv[bleh]);
									printf("raw [");
									for (bleh=0; bleh < 16; bleh++)
										printf("0x%.2x ",raw_recv[bleh]);
									printf("0x%.2x]\n",raw_recv[bleh]);
#endif
									dont_wait_for_reply = 1;
									break;
									}
								else
									{
#ifdef TRANS_DEBUG
									int bleh;
									printf("Interrupted pkt OK  [");
									for (bleh=0; norm_recv[bleh] != (unsigned char) 0xa5; bleh++)
										printf("0x%.2x ",norm_recv[bleh]);
									printf("0x%.2x]\n",norm_recv[bleh]);
#endif
									}
								if (norm_recv[5] == CMD_EMP_ACTIVE)
									{
									int	cc;
									cc = process_active(ctx, norm_recv);
									if (cc == 2)
										{
										ctx->console_mode = 1;
										ctx->enable_driver_checks = 0;
										dont_wait_for_reply = 1;
										break;
										}
									else if (cc == 0)
										{
										/* Connection was dropped. ask thread to reconnect */
#ifdef THREAD_DEBUG
										printf("Driver %s Connection dropped while sending\n",
													ctx->node_id);
#endif
										reply.rc=ECONNRESET;
										state = STATE_SEND_REPLY;
										continue;
										}
									}
								else
									{
									int dbg;
									printf("Driver %s Interrupted by 0x%.2x (l=%d)\n",
												 ctx->node_id,norm_recv[5],l);
									for (dbg = 0; dbg < l; dbg++)
										printf("0x%.2x ",norm_recv[dbg]);
									printf("0x%.2x\n",norm_recv[dbg]);
									for (dbg =0; raw_recv[dbg] != (unsigned char) 0xa5; dbg++)
										printf("0x%.2x ",raw_recv[dbg]);
									printf("0x%.2x\n",raw_recv[dbg]);
									}
								}
							recvpos = 0;
							if (full_duplex_ack)
								state = STATE_SEND_DB;
							else
								state = STATE_RECV_DA;
							}
						continue;
						}
					else if (state == STATE_SEND_REPLY)
						{
						send_driver_ipc(ctx, ctx->driver_pwrite, &reply);
						dont_wait_for_reply = 1;
						break;
						}
					}
				recv_abort = 0;
				while(!dont_wait_for_reply)
					{
					recvpos = 0;
					while(1)
						{
						unsigned char	UC_byte;
						int						rc;

						if ((rc = wait_for_byte(ctx, &UC_byte)) < 0)
							{
							reply.rc = -rc;
							recv_abort = 1;
							break;
							}
#ifdef TRANS_DEBUG
						printf("[%s] Got response byte 0x%.2x\n",ctx->node_id,UC_byte);
#endif
						if ((recvpos == 0) && (UC_byte != (unsigned char) EMP_START_CHAR))
							{
							printf("Thread %s WARNING: Received frame does not begin with 0xa0 (0x%.2x)\n",
											ctx->node_id,
											UC_byte);
							reply.rc = EIO;
							recv_abort = 1;
							break;
							}
						if (recvpos == (int) (sizeof(raw_recv)-1))
							{
							/* Resynchronize */
#ifdef DETECT_DEBUG
							printf("Driver %s calling detect_protocol as part of resync while RECV\n",
										ctx->node_id);
#endif
							if (!(rc = driver_detect_protocol(ctx, 5, 0, 1)))
								{
								reply.request = DIPCREQ_NEG;
								recv_abort = 1;
								break;
								}
							else if (rc == -EUNATCH)
								{
								ctx->console_mode = 1;
								reply.rc = EUNATCH;
								recv_abort = 1;
								break;
								}
							reply.rc = EREMOTEIO;
							recv_abort = 1;
							break;
							}
						raw_recv[recvpos++] = UC_byte;
						if (UC_byte == (unsigned char) EMP_STOP_CHAR)
							break;
						} /* End while loop for receiving packet */
					if (recv_abort)
						break;
					response_len = normalize_data(&raw_recv[0], &norm_recv[0]);
					if (!emp_validate_cksum(ctx, norm_recv, response_len, 0))
						{
						duplex_retry++;
						if (duplex_retry < 3)
							{
#ifdef TRANS_DEBUG
printf("[%s] Retrying packet due to bad cksum on response\n",ctx->node_id);
#endif
						goto retry;
							}
						printf("Giving up due to errors\n");
						rc = EIO;
						break;
						}
					memcpy(&reply.pkt_buffer, norm_recv, response_len);
    			if (norm_recv[5] != msg.cmd)
      			{
      			if (norm_recv[5] == CMD_EMP_ACTIVE)
        			{
							int	cc;
							// XXX:
							cc = process_active(ctx, norm_recv);
							if (cc == 2)
								{
								ctx->console_mode = 1;
								ctx->enable_driver_checks = 0;
								}
							else if (cc == 0)
								{
								/* Connection was dropped. ask thread to reconnect */
								printf("Driver %s Connection dropped while sending\n",
											ctx->node_id);
								reply.rc=ECONNRESET;
								recv_abort = 1;
								break;
								}
        			continue; // wait for the next reply XXX:
          		}
      			reply.rc = EPROTO;
						break;
        		}
					else
						{
#ifdef TRANS_DEBUG
						printf("[%s] RESPONSE RECEIVED cc = 0x%.2x\n",ctx->node_id,norm_recv[6]);
#endif
						if (full_duplex_ack)
							{
							if (norm_recv[6] != 0x00)
								{
								duplex_retry++;
								if (duplex_retry < 3)
									{
#ifdef TRANS_DEBUG
									printf("[%s] Retrying packet due to bad cksum on response\n",ctx->node_id);
#endif
									goto retry;
									}
								}
							}
						reply.rc = 0;
						break;
						}
      		}
#ifdef TRANS_DEBUG
	if (duplex_retry != 0)
printf("[%s] Sending back pkt rc %d, cc 0x%.2x\n",ctx->node_id,reply.rc,norm_recv[6]);
#endif
				reply.pkt_len = response_len;
				send_driver_ipc(ctx, ctx->driver_pwrite, &reply);
				}
			else if (msg.request == DIPCREQ_DET)
				{
				DRIVER_IPC	reply;

#if 0
				printf("Driver %s detecting protocol by request of thread [",
								ctx->node_id);
#endif
				/* Call to detect protocol */
				memset(&reply, 0, sizeof(reply));
				reply.request = DIPCREQ_DET;
				reply.rc = driver_detect_protocol(ctx, 5, 0, 0);
#if 0
				printf("%d]\n",reply.rc);
#endif
				send_driver_ipc(ctx, ctx->driver_pwrite, &reply);
				}
			}
		else if ((FD_ISSET(ctx->emp_sock, &read_fds)) && ((ctx->protocol_detected) || (ctx->console_mode)))
			{
			unsigned char	UC_byte;

			if ((rc = wait_for_byte(ctx, &UC_byte)) < 0)
				{
#ifdef DRIVER_DEBUG
printf("D[%s] Unable to get unsolicited byte (%s)\n",ctx->node_id, strerror(-rc));
#endif
				sprintf(log_msg,
								"0:0:LOG:Thread %s unable to get unsolicited byte (%s)",
								ctx->node_id,
								strerror(-rc));
				send_msg(ctx,ctx->thread_ipc_write, log_msg);
				continue;
				}
#ifdef DRIVER_DEBUG
printf("D[%s] got byte (0x%.2x) (recv pos %d) (full pkt %d)\n",
				ctx->node_id,
				UC_byte,
				unsolicited_recv_pos,
				full_pkt);
#endif
			if (ctx->console_mode)
				{
				/* This is a console byte... */
				if (UC_byte != (unsigned char) EMP_START_CHAR)
					{
					memset(&msg, 0, sizeof(msg));
					msg.request = DIPCREQ_CB;
					msg.byte = UC_byte;
					send_driver_ipc(ctx, ctx->driver_pwrite, &msg);
					}
				}
			if (UC_byte == (unsigned char) EMP_START_CHAR)
				{
				if (ctx->console_mode)
					{
					ctx->console_mode = 0;
					ctx->enable_driver_checks = 1;
#ifdef DETECT_DEBUG
					printf("Driver %s calling detect_protocol because 0xa0 recv'ed in console mode\n",ctx->node_id);
#endif
					if (!(rc = driver_detect_protocol(ctx, 5, 0, 1)))
						{
						memset(&msg, 0, sizeof(msg));
						msg.request = DIPCREQ_NEG;
						send_driver_ipc(ctx, ctx->driver_pwrite, &msg);
						}
					else if (rc == -EUNATCH)
						ctx->console_mode = 1;
					unsolicited_recv_pos = 0;
					full_pkt = 0;
					continue;
					}
				unsolicited_recv_pos = 0;
				full_pkt = 1;
				}
			if (full_pkt)
				{
				if (unsolicited_recv_pos == (int)(sizeof(unsolicited_pkt)-1))
					{
					/* Resynchronize */
#ifdef DETECT_DEBUG
					printf("Driver %s calling detect_protocol because resync required\n",ctx->node_id);
#endif
					if (!(rc = driver_detect_protocol(ctx, 5, 0, 1)))
						{
						memset(&msg, 0, sizeof(msg));
						msg.request = DIPCREQ_NEG;
						send_driver_ipc(ctx, ctx->driver_pwrite, &msg);
						}
					else if (rc == -EUNATCH)
						ctx->console_mode = 1;
					unsolicited_recv_pos = 0;
					full_pkt = 0;
					continue;
					}
				unsolicited_pkt[unsolicited_recv_pos++] = UC_byte;
				if (UC_byte == (unsigned char) EMP_STOP_CHAR)
					{
					unsigned char	normal_frame[EMP_MAX_MSG_LEN*2];
					int		l;
	
					unsolicited_recv_pos = 0;
					full_pkt = 0;
					l = normalize_data(&unsolicited_pkt[0], &normal_frame[0]);
					if (!emp_validate_cksum(ctx, normal_frame, l, 0))
						{
						printf("bad (3)\n");
						continue;
						}
					if (normal_frame[5] == CMD_EMP_ACTIVE)
						{
						int	cc;
						cc = process_active(ctx, normal_frame);
						if (cc == 2)
							{
							ctx->console_mode = 1;
							ctx->enable_driver_checks = 0;
							}
						else if (cc == 0)
							{	
							DRIVER_IPC conreq;
							
							memset(&conreq, 0, sizeof(conreq));
							conreq.request=DIPCREQ_CON;
							send_driver_ipc(ctx, ctx->driver_pwrite, &conreq);
							}
						}
					else
						{
#ifdef DRIVER_DEBUG
printf("D[%s] Ignoring unsolicited pkt 0x%.2x (l=%d)\n",ctx->node_id,normal_frame[5],l);
#endif
						sprintf(log_msg,
										"0:0:LOG:Driver %s ignoring unsolicited pkt 0x%.2x(l=%d)",
										ctx->node_id,
										normal_frame[5],l);
						send_msg(ctx,ctx->thread_ipc_write, log_msg);
						}
					}
				}
			continue;
			}
		else if ((FD_ISSET(ctx->emp_sock, &read_fds)) && (!ctx->protocol_detected) && (!ctx->console_mode))
			{
			sprintf(log_msg,"0:0:LOG:Driver for node %s reports unexpected serial traffic (bad cabling or hw error?)",ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			if (!ctx->session_active)
				goto do_driver_checks;
			}
		else
			{
			sprintf(log_msg,"0:0:LOG:Driver for node %s select fired for unknown entity\n",ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			}
		
		if (!ctx->session_active)
			continue;
		current = time(NULL);
		if ((current >= ctx->next_rearm) && (ctx->emp_sock != -1))
			{
			int rrc = -EAGAIN;

			while(rrc == -EAGAIN)
				{
#ifdef REARM_DEBUG
				printf("Driver %s Rearming [",ctx->node_id);
#endif
				rrc = driver_rearm(ctx);
#ifdef REARM_DEBUG
				printf("%d]\n",rrc);
#endif
				}
			ctx->next_rearm = time(NULL) + 30;
			}
do_driver_checks:
		if (ctx->enable_driver_checks)
			{
			/* See if maybe protocol is magically available (or gone) */
			if ((!ctx->protocol_detected) && (ctx->emp_sock != -1))
				{
#ifdef DETECT_DEBUG
					printf("Driver %s calling detect_protocol to see if protocol magically available\n",ctx->node_id);
#endif
				if (!(rc = driver_detect_protocol(ctx, 5, 0, 1)))
					{
					memset(&msg, 0, sizeof(msg));
					msg.request = DIPCREQ_NEG;
					send_driver_ipc(ctx, ctx->driver_pwrite, &msg);
					}
				else if (rc == -EUNATCH)
					ctx->console_mode = 1;
				}
			else if ((ctx->protocol_detected) && (ctx->emp_sock != -1))
				{
				time_t	current_time;
				int			diff_time;
				
				current_time = time(NULL);
				diff_time = (int) (current_time - ctx->last_active_timestamp);
				if (diff_time > 6)
					{
#ifdef DETECT_DEBUG
					printf("Driver %s calling detect_protocol to see if protocol magically gone\n",ctx->node_id);
#endif
					driver_detect_protocol(ctx, 5, 1, 0);
					}
				}
			}
		pthread_testcancel();
		}
}

/*
 * Function the driver can use to rearm the connection timer.
 * Returns 0 on success
 * -EAGAIN if we were interrupted by an ACTIVE packet with state 1
 * -EUNATCH if we were interrupted by an ACTIVE packet signifying console 
 * anything else is failure
 */
static int driver_rearm(CONTEXT *ctx)
{
	unsigned char		raw_send[EMP_MAX_MSG_LEN*2];
	unsigned char		norm_send[EMP_MAX_MSG_LEN*2];
	unsigned char		raw_recv[EMP_MAX_MSG_LEN*2];
	unsigned char		norm_recv[EMP_MAX_MSG_LEN*2];
	REQHDR					*hdr = (REQHDR *) &norm_send[1];
	int							len, I_c1;
	unsigned char		cksum;
	int							byte_retry_cnt = 0;
	int							I_dlen = 0, state = STATE_SEND_DB;
	int							sendpos = 0, full_interrupted_pkt =0;
	int							frame_retry_cnt = 0;
	int							recvpos = 0;
	int							response_len;
	int							full_duplex_ack= 0;

	memset(raw_send, 0, sizeof(raw_send));
	memset(norm_send, 0, sizeof(norm_send));

  hdr->rs_slave_addr= EMP_SLAVE_ADDRESS;
  hdr->net_fn       = APP_REQUEST;
  hdr->lun          = 0x00;
  hdr->cksum1       = ((0xff - hdr->rs_slave_addr - 
											(*(unsigned char *)&norm_send[2]))+1);
  hdr->rq_slave_addr= EMP_APPLICATION_SLAVE_ADDRESS;
  hdr->rqseq        = ctx->sequencenumber+=4;
  hdr->cmd          = CMD_REARM_EMP_CONNECTION_TIMER;

  cksum=0xff - hdr->rq_slave_addr - hdr->rqseq - hdr->cmd;
  for (I_c1=7;I_c1<(7+I_dlen);I_c1++)
    cksum=cksum - norm_send[I_c1];
  norm_send[7+I_dlen]= (cksum+1);

	len = unnormalize_data(&norm_send[1],
												 &raw_send[1],
												 (8+I_dlen));

	raw_send[0] = (unsigned char) EMP_START_CHAR;
	raw_send[len] = (unsigned char) EMP_STOP_CHAR;
	len++;

	recvpos = 0;
  while(1)
    {
    if (state == STATE_SEND_DB)
      {
      if (write(ctx->emp_sock, &raw_send[sendpos], 1) < 0)
				return(-EIO);
      state = STATE_RECV_DA;
      continue;
      }
    else if (state == STATE_RECV_DA)
      {
      unsigned char UC_byte;
      int           rc;
      
      if ((rc = wait_for_byte(ctx, &UC_byte)) < 0)
        {
        if (rc == -ETIMEDOUT)
          {
          byte_retry_cnt ++;
          if (byte_retry_cnt == 2)
            {
            frame_retry_cnt ++;
            if (frame_retry_cnt == 2)
              {
							printf("Driver %s re-arm: giving up\n",ctx->node_id);
							return(-EIO);
              }
            /* Retry the whole frame */
            byte_retry_cnt = 0;
            sendpos = 0;
            state = STATE_SEND_DB;
            continue;
            }
          state = STATE_SEND_DB;
          continue;
          }
				printf("Driver %s re-arm: bad rc from wait_for_byte() (%d)\n",
								ctx->node_id,
								rc);
				return(-EIO);
        }
      if (UC_byte == (unsigned char) EMP_HANDSHAKE_CHAR)
        {
        sendpos++;
        if (sendpos == len)
					{
          break;
					}
        state = STATE_SEND_DB;
        }
      else
        {
        if (UC_byte == (unsigned char) EMP_START_CHAR)
          full_interrupted_pkt = 1;
        else if (UC_byte == (unsigned char) EMP_STOP_CHAR)
          continue;
        else
          full_interrupted_pkt = 0;
        if (recvpos == (int) (sizeof(raw_recv)-1))
          {
          printf("BUFFER OVERFLOW!!!(6)\n");
          abort();
          }
        raw_recv[recvpos++] = UC_byte;
        state = STATE_RECV_DP;
        }
      continue;
      }
    else if (state == STATE_RECV_DP)
      {
      unsigned char UC_byte;
      int           rc, l;
      
      if ((rc = wait_for_byte(ctx, &UC_byte)) < 0)
        {
				printf("Driver %s re-arm: bad rc from wait_for_byte() (%d)\n",
								ctx->node_id,
								rc);
				return(-EIO);
        }
			if (UC_byte == (unsigned char) 0xa8)
				{
				full_duplex_ack = 1;
printf("[%s] Got full duplex ack\n",ctx->node_id);
				continue;
				}
      if (recvpos == (int) (sizeof(raw_recv)-1))
				{
				printf("Driver %s re-arm: recvpos > sizeof(raw_recv)\n",ctx->node_id);
				return(-EIO);
				}
      raw_recv[recvpos++] = UC_byte;
      if (UC_byte == (unsigned char) EMP_STOP_CHAR)
        {
        if (full_interrupted_pkt)
          {
          l = normalize_data(&raw_recv[0], &norm_recv[0]);
          if (!emp_validate_cksum(ctx,norm_recv, l, 0))
            {
						printf("Driver %s re-arm: bad cksum on int frame\n",ctx->node_id);
						return(-EIO);
            }
          if (norm_recv[5] == CMD_EMP_ACTIVE)
            {
            int cc;
            cc = process_active(ctx, norm_recv);
						if (cc == 1)
							{
#ifdef REARM_DEBUG
printf("Driver %s re-arm: got ping (1). Waiting for next packet\n",
				ctx->node_id);
#endif
							recvpos = 0;
							state = STATE_RECV_DA;
							continue;
							}
            else if (cc == 2)
							return(-EUNATCH);
            else if (cc == 0)
							return(-ECONNRESET);
            }
          else
            {
#ifdef REARM_DEBUG
						printf("Driver %s re-arm: int by non ACTIVE pkt (0x%.2x)\n",
										ctx->node_id,
										norm_recv[5]);
#endif
      			wait_for_byte(ctx, &UC_byte); // Swallow the ack and get out
						sleep(1); // Wait for time out the packet send on h/w side
						return(-EIO);
            }
          } 
        recvpos = 0;
				if (full_duplex_ack)
					state = STATE_SEND_DB;
				else
        	state = STATE_RECV_DA;
        }
      continue;
      }
    }

 	while(1)
    {
    recvpos = 0;
    while(1)
      {
      unsigned char UC_byte;
      int           rc;
      
      if ((rc = wait_for_byte(ctx, &UC_byte)) < 0)
				{
				printf("Driver %s re-arm: bad rc from wait_for_byte() (%d)\n",
								ctx->node_id,
								rc);
				return(-EIO);
				}
      if ((recvpos == 0) && (UC_byte != (unsigned char) EMP_START_CHAR))
				{
				printf("Driver %s re-arm: frame does not begin with 0xa0! (0x%.2x)\n",
							ctx->node_id,
							UC_byte);
				return(-EIO);
				}
      if (recvpos == (int) (sizeof(raw_recv)-1))
				{
				printf("Driver %s re-arm: recvpos > sizeof(raw_recv)\n",ctx->node_id);
				return(-EIO);
				}
      raw_recv[recvpos++] = UC_byte;
      if (UC_byte == (unsigned char) EMP_STOP_CHAR)
        break;
      } /* End while loop for receiving packet */
    response_len = normalize_data(&raw_recv[0], &norm_recv[0]);
    if (!emp_validate_cksum(ctx, norm_recv, response_len, 0))
      {
			printf("Driver %s re-arm: bad cksum on response frame\n",ctx->node_id);
			return(-EIO);
      }
    if (norm_recv[5] != CMD_REARM_EMP_CONNECTION_TIMER)
      {
      if (norm_recv[5] == CMD_EMP_ACTIVE)
        {
        int cc;
        cc = process_active(ctx, norm_recv);
				if (cc == 1)
					{
#ifdef REARM_DEBUG
printf("Driver %s re-arm: got ping (1). Waiting for next packet\n",ctx->node_id);
#endif
					recvpos = 0;
					continue;
					}
        else if (cc == 2)
					return(-EUNATCH);
        else if (cc == 0)
					return(-ECONNRESET);
        continue; // wait for the next reply
        }
			return(-EIO);
      }
    else
      break;
    }
	if (norm_recv[6] != 0x00)
		{
		printf("Driver %s re-arm: Bad CC (0x%x)\n",ctx->node_id,norm_recv[6]);
		return(-EREMOTEIO);
		}
#ifdef REARM_DEBUG
printf("Leaving");
fflush(stdout);
#endif
	return(0);
}

static int process_active(CONTEXT *ctx, unsigned char *frame)
{
	ctx->last_active_timestamp = time(NULL);
	if (frame[6] == (unsigned char) 0x00)
		ctx->session_active = 0;
	return((int) frame[6]);
}

static void driver_thread_cleanup(void *data)
{
	CONTEXT *ctx = (CONTEXT *) data;
	ctx->refcount--;
}

static int send_driver_ipc(CONTEXT *ctx, int fd, DRIVER_IPC *msg)
{
  KNOWN_UNUSED_PARAM(ctx);
	if (write(fd, (void *) msg, sizeof(DRIVER_IPC)) < 0)
		{
		printf("send_driver_ipc(): Failed (%m)\n");
		return(-errno);
		}
	return(0);
}

/*
 * Wait for protocol detection.
 * Set the protocol_detected flag in the context
 * Return 0 on success
 * Return -ETIMED out on no protocol
 * Return -EUNATCH if in console mode
 * negative errno on other error.
 * NEW: This code till try and fake out a modem if it detects AT sequences comming
 * from the hardware
 */

static int driver_detect_protocol(CONTEXT	*ctx,
													 int			timeout,
													 int			quiet_detected,
													 int			quiet_unavailable)
{
	unsigned char		frame[65];
	unsigned char		norm_frame[65];
	int							framecount = -1, l;
	struct timeval	to;
	fd_set					read_fds;
	int							rc, frc, rrc;
	unsigned char		UC_byte;
	int							non_framed_bytecount = 0;
	char						log_msg[255];
	int							modem_mode = 0;

	pthread_t				self;

	self = pthread_self();

	if (pthread_equal(self, ctx->thread))
		{
		printf("ACK! Main thread called driver_detect_protocol!\n");
		}
	while(1)
		{
		FD_ZERO(&read_fds);
		to.tv_sec = timeout;
		to.tv_usec = 0;
		FD_SET(ctx->emp_sock, &read_fds);
		if ((rc = select((ctx->emp_sock+1), &read_fds, NULL, NULL, &to)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Driver %s detect select() failure (%m)",ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			frc = -errno;
			goto unavailable;
			}
		if (!rc)
			{
			frc = -ETIMEDOUT;
			goto unavailable;
			}
		if (FD_ISSET(ctx->emp_sock, &read_fds))
			{
			if ((rrc=read(ctx->emp_sock, &UC_byte, 1)) < 0)
				{
				frc = -errno;
				goto unavailable;
				}
			if (!rrc)
				{
				sprintf(log_msg,
								"0:0:LOG:Driver %s detect read failed. Connection to hardware lost",ctx->node_id);
				send_msg(ctx,ctx->thread_ipc_write, log_msg);
				}
			if (UC_byte == (unsigned char) 0xa0)
				framecount = 0;
			if ((UC_byte == '\r') && (!modem_mode))
				{
				modem_mode = 1;
				framecount = 0;
				continue;
				}
			if (framecount >=0)
				{
				if (framecount == (int) (sizeof(frame)-1))
					{
					sprintf(log_msg,
									"0:0:LOG:Driver %s BUFFER OVERFLOW!!!(2)",ctx->node_id);
					send_msg(ctx,ctx->thread_ipc_write, log_msg);
					abort();
					}
				frame[framecount++] = UC_byte;
				if (UC_byte == (unsigned char) 0xa5)
					break;
				if ((UC_byte == '\r') && (modem_mode))
					{
					frame[framecount] = 0;
					if(!strncmp(frame,"AT",2))
						{
						sprintf(log_msg,
										"0:0:LOG:Driver %s faking modem for hardware connection",
										ctx->node_id);
						send_msg(ctx,ctx->thread_ipc_write, log_msg);
						fake_modem(ctx);
						/* Try to detect again */
						framecount = -1;
						non_framed_bytecount = 0;
						modem_mode = 0;
						continue;
						}
					else
						{
						framecount = 0;
						}
					}
				}
			else
				{
				non_framed_bytecount++;
				if (non_framed_bytecount > (EMP_MAX_MSG_LEN*2))
					goto console_mode;
				}
			}
		}
	l = normalize_data(&frame[0], &norm_frame[0]);
	if (!emp_validate_cksum(ctx, norm_frame, l, 0))
		{
		printf("bad (4)\n");
		frc = -ETIMEDOUT;
		goto unavailable;
		}
	if (!quiet_detected)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s protocol detected",
						ctx->node_id);
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		}
	ctx->protocol_detected = 1;
	ctx->last_active_timestamp = time(NULL);
	return(0);

unavailable:
	if (!quiet_unavailable)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s protocol unavailable (%s)",
					 	ctx->node_id,
					 	strerror(-frc));
		send_msg(ctx,ctx->thread_ipc_write, log_msg);
		}
	ctx->protocol_detected = 0;
	return(frc);
console_mode:
	sprintf(log_msg,
					"0:0:LOG:Thread %s protocol unavailable (%s)",
				 	ctx->node_id,
				 	strerror(EUNATCH));
	send_msg(ctx,ctx->thread_ipc_write, log_msg);
	ctx->protocol_detected = 0;
	return(-EUNATCH);
}

static int fake_modem(CONTEXT *ctx)
{
	struct timeval	to;
	fd_set					read_fds;
	char						data[255];
	int							datapos = 0;
	int							rc;
	int							capture = 0;
	char						mybyte;
	char						log_msg[255];

	sprintf(log_msg,
					"0:0:LOG:Driver %s modem faking commencing",
				 	ctx->node_id);
	send_msg(ctx,ctx->thread_ipc_write, log_msg);
	while(1)
		{
		pthread_testcancel();
		FD_ZERO(&read_fds);
		to.tv_sec = 5;
		to.tv_usec = 0;
		FD_SET(ctx->emp_sock, &read_fds);
		if ((rc = select((ctx->emp_sock+1), &read_fds, NULL, NULL, &to)) < 0)
			return(-errno);
		if (!rc)
			{
			sprintf(log_msg,
							"0:0:LOG:Driver %s sending 'RING'",
				 			ctx->node_id);
			send_msg(ctx,ctx->thread_ipc_write, log_msg);
			write(ctx->emp_sock, "RING\r",5);
			}
		if (FD_ISSET(ctx->emp_sock, &read_fds))
			{
			if ((rc=read(ctx->emp_sock, &mybyte, 1))<0)
				return(-errno);
			if ((mybyte == '\r') && (!capture))
				{
				capture = 1;
				continue;
				}
			else if (capture)
				{
				data[datapos++] = mybyte;
				if (mybyte == '\r')
					{
					data[(datapos-1)] = 0;
					if (!strcasecmp(data, "ATA"))
						{
						sprintf(log_msg,
										"0:0:LOG:Driver %s got 'ATA'; sending 'CONNECT'",
				 						ctx->node_id);
						send_msg(ctx,ctx->thread_ipc_write, log_msg);
						write(ctx->emp_sock, "CONNECT\r",8);
						sprintf(log_msg,
										"0:0:LOG:Driver %s modem faking complete",
				 						ctx->node_id);
						return(0);
						}
					else if (!strncasecmp(data, "AT", 2))
						{
						sprintf(log_msg,
										"0:0:LOG:Driver %s got 'AT' sending 'OK'",
				 						ctx->node_id);
						send_msg(ctx,ctx->thread_ipc_write, log_msg);
						write(ctx->emp_sock, "OK\r",3);
						}
					memset(data, 0, sizeof(data));
					datapos = 0;
					}
				}
			}
		}
}

static int emp_validate_cksum(CONTEXT *ctx, unsigned char *frame, int framelen, int quiet)
{
  char          log_msg[255];
  int           i;
  unsigned char cksum_acc = 0x00;
  char          tmp[16];
  int           j;

  /* First check the header checksum */
  if ((((unsigned char) 0xFF - frame[0] - frame[1])+1) != frame[2])
    {
    if (!quiet)
      {
      sprintf(log_msg,
              "0:0:LOG:Thread %s received corrupt EMP header (us 0x%.2x, them 0x%.2x)",
              ctx->node_id,
              cksum_acc,
              frame[2]);
      send_msg(ctx,ctx->thread_ipc_write, log_msg);
#ifdef TRANS_DEBUG
printf(
      "Thread %s received corrupt EMP header (us 0x%.2x, them 0x%.2x)",
      ctx->node_id,
      cksum_acc,
      frame[2]);
#endif
      sprintf(log_msg,"0:0:LOG:Thread %s data ",ctx->node_id);
#ifdef TRANS_DEBUG
printf("Thread %s data ",ctx->node_id);
#endif
      for (j = 0; j < framelen; j++)
        {
#ifdef TRANS_DEBUG
printf("0x%.2x ",frame[j]);
#endif
        sprintf(tmp,"%.2x ",frame[j]);
        strcat(log_msg,tmp);
        }
#ifdef TRANS_DEBUG
printf("0x%.2x\n",frame[j]);
#endif
      sprintf(tmp,"%.2x ",frame[j]);
      strcat(log_msg,tmp);
      send_msg(ctx,ctx->thread_ipc_write, log_msg);
      }
    return(0);
    }

  /* Now check the data section */
  cksum_acc = 0xFF;
  for (i=3; i < (framelen-1); i++)
    cksum_acc -=frame[i];
  cksum_acc +=1;
  if (cksum_acc != frame[i])
    {
    if (!quiet)
      {
      sprintf(log_msg,
              "0:0:LOG:Thread %s received corrupt EMP data field (us 0x%.2x, them 0x%.2x)",
              ctx->node_id,
              cksum_acc,
              frame[i]);
      send_msg(ctx,ctx->thread_ipc_write, log_msg);
      sprintf(log_msg,"0:0:LOG:Thread %s data ",ctx->node_id);
      for (j = 0; j < framelen; j++)
        {
        sprintf(tmp,"%.2x ",frame[j]);
        strcat(log_msg,tmp);
        }
      sprintf(tmp,"%.2x ",frame[j]);
      strcat(log_msg,tmp);
      send_msg(ctx,ctx->thread_ipc_write, log_msg);
      }
    return(0);
    }
#if 0
  printf("cksum ok (we got 0x%.2x, pkt is 0x%.2x)\n",
          cksum_acc, frame[i]);
#endif
  return(1);
}

/*
 *  * Wait for up to 1s for byte to arrive on serial port
 *   * Return 0 and the byte in bytep if successfull
 *    */
static int wait_for_byte(CONTEXT *ctx, unsigned char *bytep)
{
  struct timeval  to;
  fd_set          read_fds;
  int             rc;
#ifdef BYTE_RECV_DEBUG
  pthread_t       caller;

  caller = pthread_self();
  printf("wait_for_byte(): Node %s called by %s\n",
         ctx->node_id,
         (caller == ctx->driver_thread ? "DRIVER" : "THREAD"));
#endif
  FD_ZERO(&read_fds);
  to.tv_sec = 1;
  to.tv_usec = 0;

  FD_SET(ctx->emp_sock, &read_fds);
  if ((rc = select((ctx->emp_sock+1), &read_fds, NULL, NULL, &to)) < 0)
    {
#ifdef BYTE_RECV_DEBUG
    printf("wait_for_byte(): select failure (%m)\n");
#endif
    return(-errno);
    }
  if (!rc)
    {
#ifdef BYTE_RECV_DEBUG
    printf("wait_for_byte(): timeout\n");
#endif
    return(-ETIMEDOUT);
    }
  if (!FD_ISSET(ctx->emp_sock, &read_fds))
    {
#ifdef BYTE_RECV_DEBUG
    printf("wait_for_byte(): FD is not set??\n");
#endif
    return(-EIO);
    }
  else
    {
    if (read(ctx->emp_sock, (void *) bytep, 1) < 0)
      {
#ifdef BYTE_RECV_DEBUG
      printf("wait_for_byte(): read failure (%m)\n");
#endif
      return(-errno);
      }
    }
  return(0);
}

