/*
 * This program is original work by San Mehat <nettwerk@valinux.com> and
 * Zac Sprackett <zac@sprackett.com>.
 * This program is distributed under the GNU Lesser General Public License
 * (LGPL).
 *
 * Copyright (C) 1999, San Mehat, Zac Sprackett and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <sys/time.h>
#include <errno.h>
#include <signal.h>
#include <libloose.h>
#include <libloose_private.h>
#include <util.h>

static void lm_dispatch_timers(void);

static List           *lm_fd_watch;
static List           *lm_timer;
static int						I_fd = -1;
static int						has_registered = 0;
extern int						errno;

#define TRUE	1
#define FALSE	!TRUE

int lm_init(void)
{
	lm_fd_watch = (List *)malloc(sizeof(List));
	lm_timer = (List *)malloc(sizeof(List));
	list_init(lm_fd_watch,destroy_fd_watcher);
	list_init(lm_timer,destroy_timer);

	signal(SIGINT, SIG_IGN);
	return(0);
}

int lm_nexxus_disconnect(void)
{
	if (I_fd == -1)
		{
		printf("[lm_nexxus_disconnect] Not connected to Nexxus\n");
		return(-1);
		}
	close(I_fd);
	I_fd = -1;
	return(0);
}

int lm_nexxus_connect(void)
{
	struct sockaddr_un	sock_un;
	pid_t								pid;
	char								pid_msg[32];

	if (I_fd != -1)
		{
		printf("[lm_nexxus_connect] Allready connected to Nexxus\n");
		return(-1);
		}
	if ((I_fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
		{
		printf("[lm_nexxus_connect] Unable to create Nexxus socket (%m)\n");
		return(-1);
		}
	memset(&sock_un, 0, sizeof(sock_un));

	sock_un.sun_family = AF_UNIX;
	strcpy(sock_un.sun_path, "/tmp/vacm");

	if (connect(I_fd, (struct sockaddr *) &sock_un, sizeof(struct sockaddr)) < 0)
		{
		printf("[lm_nexxus_connect] Unable to connect to Nexxus (%m)\n");
		return(-1);
		}
	pid = getpid();
	sprintf(pid_msg, "0:0:PID:%d",pid);
	if (lm_send_to_nexxus(pid_msg) < 0)
		{
		printf("[lm_nexxus_connect] Unable to send PID (%m)\n");
		return(-1);
		}
	return(0);
}

void lm_main_loop(void (*config_cb) (__uint32_t, int, char *),
								  void (*deinit_cb) (__uint32_t, int, char *),
								  void (*discon_cb) (__uint32_t, int, char *),
								  void (*ipc_cb) (char *, __uint32_t, int, char *))
{
	char						node_id[64], saved_buffer[255], buffer[255];
	fd_set					read_fds;
	char						*bword, *cmd, *module, *p;
	__uint32_t			job_id;
	int							max, rc, client_fd;
	struct timeval	to;
	List_Element    *fd_watch_scan;

	while(1)
		{
		FD_ZERO(&read_fds);
		FD_SET(I_fd, &read_fds);
		max = 0;
		to.tv_sec = 1;
		to.tv_usec = 0;
		
		if (I_fd > max)
			max = I_fd;

		/* Loop through our watchers */
		fd_watch_scan = list_head(lm_fd_watch);
		while(fd_watch_scan)
			{
			LM_FD_WATCHER *watcher = list_data(fd_watch_scan);
			FD_SET(watcher->I_fd, &read_fds);

			if(watcher->I_fd > max)
				max = watcher->I_fd;

			fd_watch_scan = fd_watch_scan->next;
			}

		if ((rc = select((max+1), &read_fds, NULL, NULL, &to))<0)
			{
			if ((errno == EINTR) || (errno == EPIPE))
				continue;
			printf("[lm_main_loop] select() failure (%m)\n");
			exit(-1);
			}
		if (!rc)
			{
			lm_dispatch_timers();
			continue;
			}
		if (FD_ISSET(I_fd, &read_fds))
			{
			__uint32_t	length;
			if ((rc =read(I_fd, &length, sizeof(length))) < 0)
				{
				printf("[lm_main_loop] Unable to read length from Nexxus (%m)\n");
				deinit_cb(0, 0, NULL);
				}
			length = ntohl(length);

			if ((rc =read(I_fd, buffer, length)) < 0)
				{
				printf("[lm_main_loop] Unable to read data from Nexxus (%m)\n");
				perror("read");
				deinit_cb(0, 0, NULL);
				}
			if (!rc)
				{
				printf("[lm_main_loop] Read of length 0 from Nexxus (%m)\n");
				deinit_cb(0, 0, NULL);
				}
			memcpy(saved_buffer, buffer, (strlen(buffer)+1));

			if (!(bword = strtok(buffer, ":")))
				goto error_continue;
			job_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);

			if (!(bword = strtok(NULL, ":")))
				goto error_continue;
			client_fd = atoi(bword);

			if (!(cmd = strtok(NULL, ":")))
				goto error_continue;
			p = &saved_buffer[((cmd - &buffer[0]) + strlen(cmd)+1)];
			if (!(strcasecmp(cmd, "CFG")))
				{
				 // printf("cfg <%s>\n",saved_buffer); 
				config_cb(job_id, client_fd, p);
				}
			else if (!(strcasecmp(cmd, "DEINIT")))
				deinit_cb(job_id, client_fd, p);
			else if (!(strcasecmp(cmd, "DISCON")))
				discon_cb(job_id, client_fd, p);
			else if (!(strcasecmp(cmd, "IPC")))
				{
				if (!(bword = strtok(NULL,":")))
					goto error_continue;
				strncpy(node_id, bword, (sizeof(node_id)-1));
				if (!(module = strtok(NULL,":")))
					goto error_continue;
				p = &saved_buffer[((module - &buffer[0]) + strlen(module)+1)];
				ipc_cb(node_id, job_id, client_fd, p);
				}
			else
				{
				printf("[libloose] Error! Invalid CMD specified in packet (%s)\n",cmd);
				goto error_continue;
				}
			}
		/* Loop through our watchers */
		fd_watch_scan = list_head(lm_fd_watch);
		while(fd_watch_scan)
			{
			LM_FD_WATCHER *watcher = list_data(fd_watch_scan);
			if (FD_ISSET(watcher->I_fd, &read_fds))
				watcher->cb(watcher->I_fd,watcher->arg);

			fd_watch_scan = fd_watch_scan->next;
			}
		lm_dispatch_timers();
		continue;
error_continue:
		printf("[lm_main_loop] Malformatted data from Nexxus received\n");
		printf("[lm_main_loop] Data was <%s>\n",saved_buffer);
		}
}

int lm_log(char *msg)
{
	char	logmsg[255];

	if (!has_registered)
		{
		printf("[lm_log] Module not registered\n");
		return(-1);
		}
	sprintf(logmsg,"0:0:LOG:%s", msg);
	return(lm_send_to_nexxus(logmsg));
}

int lm_register(char	*short_name,
								char	*long_name,
								char	*desc,
								char	*author,
								char	*ipc_tag,
								int		major_version,
								int		minor_version)
{
	char	msg[255];

	if (has_registered)
		{
		printf("[lm_register] Module already registered\n");
		return(-1);
		}
	else
		has_registered++;

	sprintf(msg,"0:0:REG:%s:%s:%s:%s:%d:%d:%s",
					short_name,
					long_name,
					desc,
					author,
					major_version,
					minor_version,
					ipc_tag);
	return(lm_send_to_nexxus(msg));
}

int lm_send_to_nexxus(char *msg)
{
	__uint32_t	length;
	__uint32_t	network_length;

	if (I_fd == -1)
		{
		printf("[lm_send_to_nexxus] Not connected to Nexxus\n");
		return(-1);
		}
	length = (strlen(msg)+1);
	network_length = htonl(length);
	if (write(I_fd, &network_length, sizeof(network_length)) < 0)
		{
		printf("[lm_send_to_nexxus] Error writing length (%m)\n");
		return(-1);
		}
	if (write(I_fd, msg, length) < 0)
		{
		printf("[lm_send_to_nexxus] Error writing data (%m)\n");
		return(-1);
		}
	return(0);
}

int lm_watch_fd(int		I_fd, 
								void	(*cb) (int, void *),
								void	*arg)
{
	LM_FD_WATCHER *watcher = NULL;

	watcher = (LM_FD_WATCHER *)malloc(sizeof(LM_FD_WATCHER));

	if(!watcher)
		return(-1);

	watcher->I_fd  = I_fd;
	watcher->cb    = cb;
	watcher->arg   = arg;

	if((list_ins_next(lm_fd_watch, list_tail(lm_fd_watch),(const void *)watcher))<0)
		{
		destroy_fd_watcher((void *)watcher);
		return(-1);
		}
	
	return(0);
}

int lm_unwatch_fd(int	I_fd)
{
	List_Element *fd_watch_scan;

	/* Loop through our watchers */
	fd_watch_scan = list_head(lm_fd_watch);
	while(fd_watch_scan)
		{
		LM_FD_WATCHER *watcher = list_data(fd_watch_scan);
		if (watcher->I_fd == I_fd)
			{
			void *old = NULL;
			list_rem(lm_fd_watch,fd_watch_scan,&old,NULL);
			destroy_fd_watcher(old);
			return(0);
			}

		fd_watch_scan = fd_watch_scan->next;
		}

	printf("[lm_unwatch_fd] FD %d not found on watchlist\n",I_fd);
	return(-1);
}

static void lm_dispatch_timers(void)
{
	unsigned int	current_time;
	List_Element *timer_scan;

	current_time = time(NULL);

	/* Loop through our timers */
	timer_scan = list_head(lm_timer);
	while(timer_scan)
		{
		LM_TIMER *timer = list_data(timer_scan);
		if (current_time >= timer->alarm_time)
			{
			if (timer->cb(timer->arg) == TRUE)
				timer->alarm_time = time(NULL) + timer->freq;
			else
				{
				void *old = NULL;
				list_rem(lm_timer,timer_scan,&old,&timer_scan);
				destroy_timer(old);
				continue;
				}
			}
		timer_scan = timer_scan->next;
		}
}

int lm_timer_add(int timeout,
								 int (*cb) (void *),
								 void *arg)
{
	LM_TIMER *timer = NULL;
	timer = (LM_TIMER *)malloc(sizeof(LM_TIMER));

	if (!timer)
		return(-1);

	timer->freq       = timeout;
	timer->alarm_time = time(NULL) + timeout;
	timer->cb         = cb;
	timer->arg        = arg;
	timer->tag        = create_timer_tag();

	if((list_ins_next(lm_timer, list_tail(lm_timer),(const void *)timer))<0)
		{
		destroy_timer((void *)timer);
		return(-1);
		}

	return(timer->tag);
}

int	lm_timer_remove(int	tag)
{
	List_Element *timer_scan;

	/* Loop through our timers */
	timer_scan = list_head(lm_timer);
	while(timer_scan)
		{
		LM_TIMER *timer = list_data(timer_scan);
		if (timer->tag == tag)
			{
			void *old = NULL;
			list_rem(lm_timer,timer_scan,&old,NULL);
			destroy_timer(old);
			return(0);
			}
		timer_scan = timer_scan->next;
		}

	printf("[lm_timer_remove] Timer %d not found on Timerlist\n",tag);
	return(-1);
}

void destroy_fd_watcher(void *data)
{
	LM_FD_WATCHER *watcher = (LM_FD_WATCHER *)data;
	free(watcher);
}

void destroy_timer(void *data)
{
	LM_TIMER *timer = (LM_TIMER *)data;
	free(timer);
}

int create_timer_tag()
{
	List_Element *timer_scan;
	int new_tag = 0;

	timer_scan = list_head(lm_timer);
	while(timer_scan)
		{
		LM_TIMER *timer = list_data(timer_scan);
		if(timer->tag > new_tag)
			new_tag = timer->tag;
		timer_scan = timer_scan->next;
		}
	new_tag++;

	return(new_tag);
}
