
/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 */

#include <config.h>

#include <stdio.h>
#include <sys/socket.h>
#include <signal.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <stdlib.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <crypt.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <fnmatch.h>
#include <ctype.h>
#include <nexxus.h>
#include <prototypes.h>

NEXXUS_USER	*user_head	= NULL;
NEXXUS_USER	*user_tail	= NULL;

/*
 * Create and link in a new Nexxus User structure
 */
NEXXUS_USER	*user_create(char	*user_name)
{
	NEXXUS_USER	*new;
	char				tmp_str[255];

	sprintf(tmp_str,"nexxus user structure <%s>",user_name);
	if (!(new = (NEXXUS_USER *) nettmcs_alloc (sizeof(NEXXUS_USER),
																						 tmp_str)))
		{
		nexxus_log_printf(NULL,
													"Unable to allocate memory for Nexxus User\n");
		return(NULL);
		}
	memset(new, 0, sizeof(NEXXUS_USER));

	if (!user_tail)
		user_tail = user_head = new;
	else
		{
		user_tail->next = new;
		new->prev = user_tail;
		}
	user_tail = new;
	strcpy(new->user_name, user_name);
	return(new);
}

/*
 * Destroys a user and all his ACL stuff
 */
void user_destroy(NEXXUS_USER	*user)
{
	USER_MODULE_ACL		*mod_acl;
	USER_ADDRESS_RULE	*addr_acl;

	if (user == user_head)
		{
		user_head = user_head->next;
		if (user_head)
			user_head->prev = NULL;
		else
			user_tail = NULL;
		}
	else if (user == user_tail)
		{
		user_tail = user_tail->prev;
		if (user_tail)
			user_tail->next = NULL;
		else
			user_head = NULL;
		}
	else
		{
		user->next->prev = user->prev;
		user->prev->next = user->next;
		}
	/*
	 * Remove ACL stuff
	 */
	if (user->mod_acl)
		{
		mod_acl = user->mod_acl;
		while(mod_acl)
			{
			user_module_acl_destroy(user, mod_acl);
			mod_acl = user->mod_acl;
			}
		}
	if (user->addr_acl)
		{
		addr_acl = user->addr_acl;
		while(addr_acl)
			{
			user_address_acl_rule_destroy(user, addr_acl);
			addr_acl = user->addr_acl;
			}
		}
	nettmcs_free(user);
}

/*
 * lookup 'user_name' and check it's password against 'password'
 * Returns 1 if ok, 0 if not
 */

int	authenticate_user(char *user_name, char *password)
{
	NEXXUS_USER	*user;
	char				*crypted;

	if (!(user = user_lookup(user_name)))
		return(0);  /* Username doesnt exist */
	crypted = crypt(password, "HI");
	if (strcmp(crypted, user->password))
		return(0); /* wrong password */
	return(1); /* OK! */
}

int	user_exists(char *user_name)
{
	NEXXUS_USER	*scan;

	scan = user_head;
	while(scan)
		{
		if (!strcmp(scan->user_name, user_name))
			return(1);
		scan = scan->next;
		}
	return(0);
}

NEXXUS_USER	*user_lookup(char	*user_name)
{
	NEXXUS_USER	*scan;

	scan = user_head;
	while(scan)
		{
		if (!strcmp(scan->user_name, user_name))
			return(scan);
		scan = scan->next;
		}
	return(NULL);
}

/* Lookup the ACL for a module and a user */
USER_MODULE_ACL *user_module_acl_lookup(NEXXUS_USER	*user,
																			 char					*module)
{
	USER_MODULE_ACL	*scan;

	scan = user->mod_acl;
	while(scan)
		{
		if (!strcasecmp(scan->module_ipc_tag, module))
			return(scan);
		scan = scan->next;
		}
	return(NULL);
}

void user_module_acl_destroy(NEXXUS_USER	*user, USER_MODULE_ACL *acl)
{
	MODULE_RULE	*rules;

	if (acl == user->mod_acl)
		{
		user->mod_acl = acl->next;
		if (user->mod_acl)
			user->mod_acl->prev = NULL;
		}
	else if (acl->next == NULL)
		acl->prev->next = NULL;
	else
		{
		acl->next->prev = acl->prev;
		acl->prev->next = acl->next;
		}
	rules = acl->rules;
	while(rules)
		{
		user_module_acl_rule_destroy(acl, rules);
		rules = acl->rules;
		}
	nettmcs_free(acl);
}

void user_module_acl_rule_destroy(USER_MODULE_ACL	*acl, MODULE_RULE *rule)
{
	if (rule == acl->rules)
		{
		acl->rules = rule->next;
		if (acl->rules)
			acl->rules->prev = NULL;
		}
	else if (rule->next == NULL)
		rule->prev->next = NULL;
	else
		{
		rule->next->prev = rule->prev;
		rule->prev->next = rule->next;
		}
	nettmcs_free(rule);
}

void user_address_acl_rule_destroy(NEXXUS_USER	*user, USER_ADDRESS_RULE *rule)
{
	if (rule == user->addr_acl)
		{
		user->addr_acl = rule->next;
		if (user->addr_acl)
			user->addr_acl->prev = NULL;
		}
	else if (rule->next == NULL)
		rule->prev->next = NULL;
	else
		{
		rule->next->prev = rule->prev;
		rule->prev->next = rule->next;
		}
	nettmcs_free(rule);
}

/*
 * Create the module ACL entry for a module, given a user.
 * Links the ACL entry into the user structure properly
 */
USER_MODULE_ACL	*user_module_acl_create(NEXXUS_USER *user,
																				char				*module)
{
	USER_MODULE_ACL	*new;
	USER_MODULE_ACL	*scan;
	char						tmp_str[255];

	sprintf(tmp_str,"user mod acl structure <%s>",user->user_name);
	if (!(new = (USER_MODULE_ACL *) nettmcs_alloc (sizeof(USER_MODULE_ACL),
																								 tmp_str)))
		{
		nexxus_log_printf(NULL,
													"Unable to allocate memory for Module ACL\n");
		return(NULL);
		}
	memset(new, 0, sizeof(USER_MODULE_ACL));

	if (!user->mod_acl)
		user->mod_acl = new;
	else
		{
		scan = user->mod_acl;
		while(scan->next)
			scan = scan->next;
		scan->next = new;
		new->prev = scan;
		}
	strcpy(new->module_ipc_tag, module);
	return(new);
}

/*
 * XXX: should find ambiguous and conflicting rules.. better use of sigbits..
 */
USER_ADDRESS_RULE *user_address_acl_rule_lookup(NEXXUS_USER			*user,
														 										int							policy,
														 										struct in_addr	*addr,
														 										int							sigbits)
{
	USER_ADDRESS_RULE	*scan;

	scan = user->addr_acl;
	while(scan)
		{
		if ((policy == scan->policy) &&
				(addr->s_addr == scan->ipv4_address.s_addr) &&
				(sigbits == scan->ipv4_sigbits))
			return(scan);

		scan = scan->next;
		}
	return(NULL);
}

/* XXX: This should be smart enough to figure out different glob policies
 * that match.. but not today..
 * XXX: should also find if the passed in rule conflicts with a previous rule
 */
MODULE_RULE *user_module_acl_rule_lookup(NEXXUS_USER	*user,
												 			  				 char					*module,
												 								 int					policy,
												 								 char					*glob)
{
	USER_MODULE_ACL	*mod_scan;
	MODULE_RULE			*rule_scan;

	mod_scan = user->mod_acl;
	while(mod_scan)
		{
		if (!strcasecmp(module, mod_scan->module_ipc_tag))
			break;
		mod_scan = mod_scan->next;
		}
	if (!mod_scan)
		return(NULL); 

	rule_scan = mod_scan->rules;
	while(rule_scan)
		{
		if ((!strcasecmp(rule_scan->ipc_glob, glob)) &&
				(policy == rule_scan->policy))
			return(rule_scan);
		rule_scan = rule_scan->next;
		}
	return(NULL);
}

MODULE_RULE	*user_module_acl_rule_create(USER_MODULE_ACL	*acl,
																				 int							policy,
																				 char							*glob)
{
	MODULE_RULE	*new;
	MODULE_RULE	*scan;
	int					i;

	for (i=0; i < (int) strlen(glob); i++)
		{
		if (isalpha(glob[i]))
			glob[i] = toupper(glob[i]);
		}

	if (!(new = (MODULE_RULE *) nettmcs_alloc (sizeof(MODULE_RULE),
																						"module acl cmd structure")))
		{
		nexxus_log_printf(NULL,
													"Unable to allocate memory for Module CMD ACL\n");
		return(NULL);
		}
	memset(new, 0, sizeof(MODULE_RULE));

	if (!acl->rules)
		acl->rules = new;
	else 
		{
		scan = acl->rules;
		while(scan->next)
			scan = scan->next;
		scan->next = new;
		new->prev = scan;
		}
	strcpy(new->ipc_glob, glob);
	new->policy = policy;
	return(new);
}

USER_ADDRESS_RULE	*user_address_acl_rule_create(NEXXUS_USER			*user,
																					 			struct in_addr	*addr,
																					 			int							sigbits,
																					 			int							policy)
{
	USER_ADDRESS_RULE	*new;
	USER_ADDRESS_RULE	*scan;
	char							tmp_str[255];

	sprintf(tmp_str,"user addr acl structure <%s>",user->user_name);

	if (!(new = (USER_ADDRESS_RULE *) nettmcs_alloc (sizeof(USER_ADDRESS_RULE),
																									 tmp_str)))
		{
		nexxus_log_printf(NULL,
													"Unable to allocate memory for Address ACL\n");
		return(NULL);
		}
	memset(new, 0, sizeof(USER_ADDRESS_RULE));

	if (!user->addr_acl)
		user->addr_acl = new;
	else
		{
		scan = user->addr_acl;
		while(scan->next)
			scan = scan->next;
		scan->next = new;
		new->prev = scan;
		}
	memcpy(&new->ipv4_address, addr, sizeof(struct in_addr));
	new->policy = policy;
	new->ipv4_sigbits = sigbits;

	return(new);
}

int verify_ipc_permissions(NEXXUS_USER *user, char *module, char *cmd_request)
{
	USER_MODULE_ACL	*acl;
	MODULE_RULE			*rule;
	int							i;

#if 0
	printf("Verifying permissions for user %s to execute %s on module %s\n",
					user->user_name,
					cmd_request,
					module);
#endif
	for (i=0; i < (int) strlen(cmd_request); i++)
		{
		if (isalpha(cmd_request[i]))
			cmd_request[i] = toupper(cmd_request[i]);
		}

	acl = user->mod_acl;
	while(acl)
		{
		if (!strcasecmp(module, acl->module_ipc_tag))
			break;
		acl = acl ->next;
		}
	if (!acl)
		goto no_rule_found;
	rule = acl->rules;
	while(rule)
		{
		if (!(fnmatch(rule->ipc_glob, cmd_request, 0)))
			{
			if (rule->policy == POLICY_ALLOW)
				goto allow;
			else
				goto deny;
			}
			break;
		rule = rule->next;
		}

no_rule_found:
#if 0
	printf("No specific rule. Using default rule %s\n",
			(user->default_module_acl_policy == POLICY_ALLOW ? "ALLOW" : "DENY"));
#endif
	if (user->default_module_acl_policy == POLICY_ALLOW)
		goto allow;
	else
		goto deny;

allow:
	return(1);
deny:
	return(0);
}

int verify_address_permissions(NEXXUS_USER *user, struct in_addr *addr)
{
	USER_ADDRESS_RULE	*rule_scan;
	unsigned int			netmask;
	int								num_matches = 0;
	int								most_specific_match=0;
	int								most_specific_match_policy=POLICY_DENY;
	
#if 0
	printf("Verifying permissions for user %s to connect from %s\n",
					user->user_name,
					inet_ntoa(*addr));
#endif
	/* Look for a rule that matches this address */
	rule_scan = user->addr_acl;
	while(rule_scan)
		{
		netmask = (htonl(0xffffffff << (32 - rule_scan->ipv4_sigbits)));
		if ((addr->s_addr & netmask) == (rule_scan->ipv4_address.s_addr & netmask))
			{
#if 0
			printf("netmask match found on rule %p\n",rule_scan);
#endif
			num_matches++;
			if (rule_scan->ipv4_sigbits > most_specific_match)
				{
				most_specific_match = rule_scan->ipv4_sigbits;
				most_specific_match_policy = rule_scan->policy;
				}
			}
		rule_scan = rule_scan -> next;
		}
#if 0
	printf("Total matches found  %d\n",num_matches);
#endif
	if (num_matches)
		{
#if 0
		printf("Most specific match  %d\n",most_specific_match);
		printf("Most specific policy %d\n",most_specific_match_policy);
#endif
		if (most_specific_match_policy == POLICY_ALLOW)
			goto allow;
		else
			goto deny;
		}
	else
		{
#if 0
		printf("No specific rule. Using default rule %s\n",
				(user->default_address_acl_policy == POLICY_ALLOW ? "ALLOW" : "DENY"));
#endif
		if (user->default_address_acl_policy == POLICY_ALLOW)
			goto allow;
		else
			goto deny;
		}
allow:
	return(1);
deny:
	return(0);
}

void verify_users()
{
	NEXXUS_USER				*default_user;
	USER_ADDRESS_RULE	*addr_acl;
	struct in_addr		addr;

	if (user_head)
		return;
	nexxus_log_printf(NULL,
												"No users. Creating dflt user 'blum' w/ pass 'frub'\n");
	nexxus_log_printf(NULL,
												"User will be added to group 'default'\n");
	nexxus_log_printf(NULL,
												"YOU SHOULD RENAME THIS USER AND SET A NEW PASSWD\n");

	if (!(default_user = user_create("blum")))
		{
		nexxus_log_printf(NULL,"Unable to create default user\n");
		return;
		}
	strcpy(default_user->password,crypt("frub","HI"));
	default_user->default_address_acl_policy = POLICY_DENY;
	default_user->default_module_acl_policy = POLICY_ALLOW;

	addr.s_addr = inet_addr("127.0.0.1");
	if (!(addr_acl = user_address_acl_rule_create(default_user,
																					 			&addr,
																					 			32,
																					 			POLICY_ALLOW)))
		{
		nexxus_log_printf(NULL,"Unable to create default user ACL policy\n");
		return;
		}
	user_group_member_add(default_user, "default");
	configuration_save();
	return;
}

/*
 *--------------------------------------------------
 *User group support
 */

/*
 * Check if a user is a member of a group. Return group struct on success
 */
USER_GROUP	*user_group_user_is_member(NEXXUS_USER	*user, char	*group_name)
{
	USER_GROUP	*scan;

	scan = user->groups;
	while(scan)
		{
		if (!strcasecmp(group_name, scan->group_name))
			return(scan);
		scan = scan->next;
		}
	return(NULL);
}

/*
 * Create a group structure and link it to user
 */
USER_GROUP	*user_group_member_add(NEXXUS_USER	*user,
															 		 char					*group_name)
{
	USER_GROUP	*new;
	USER_GROUP	*scan;

	new = (USER_GROUP *) nettmcs_alloc(sizeof(USER_GROUP),"user group structure");
	memset(new, 0, sizeof(USER_GROUP));
	strncpy(new->group_name, group_name, GROUP_NAME_MAXLENGTH);

	if (!user->groups)
		user->groups = new;
	else
		{
		scan = user->groups;
		while(scan->next)
			scan = scan->next;
		scan->next = new;
		new->prev = scan;
		}
	
	return(new);
}

/*
 * Remove user specified from group
 */
void user_group_member_remove(NEXXUS_USER	*user,
											 		 		char				*group_name)
{
	USER_GROUP	*scan;

	scan = user->groups;
	while(scan)
		{
		if (!strcasecmp(group_name, scan->group_name))
			{
			if (scan == user->groups)
				{
				user->groups = scan->next;
				if (user->groups)
					user->groups->prev = NULL;
				}
			else if (scan->next == NULL)
				scan->prev->next = NULL;
			else
				{
				scan->prev->next = scan->next;
				scan->next->prev = scan->prev;
				}
			break;
			}
		scan = scan->next;
		}
	nettmcs_free(scan);
}
