/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Notes: This is now a stand-alone run-time linkable module for VACM
 * Symbols beginning with 'api' denote a module harness required
 * symbol which should not be renamed. Renaming an 'api' symbol
 * will cause undesired operation, and program failure.
 *
 */

#ifndef _VASH_H
#define _VASH_H

#ifndef HAVE_LIBREADLINE
typedef int Function ();
typedef char **CPPFunction ();
#define whitespace(c) (((c) == ' ') || ((c) == '\t'))
#endif

struct handle
	{
	char	*name;
	void	*handle;
  struct handle *next;
	};

struct alias
	{
	char *alias;
	char *expanded;
  struct alias *next;
	};

typedef struct
	{
	char *name;
	Function *callback;
	char *help;
	} COMMAND;

static int handle_user_input(char *SZ_line);
#ifdef HAVE_LIBREADLINE
static void rl_handle_user_input (char *line);
static char **rl_tab_completion(char *line,int start);
static char *rl_get_full_command_matches(char *word, int state);
static char *rl_get_command_matches(char *word, int state);
static char *rl_get_alias_matches(char *word, int state);
static char *rl_get_nexxus_matches(char *nexxus, int state);
#else 
static void no_rl_callback_read_char(void);
#endif

static struct alias *create_alias(void);
static void destroy_alias(struct alias *ptr, struct alias **scan_ptr);
static struct handle *create_handle(void);
static void destroy_handle(struct handle *ptr, struct handle **scan_ptr);

static int do_connect(char *nexxus, char *login, char *pass);
static int do_exit(int exitval);

/* Callback Functions */
static int alias_cb(char *arg);
static int connect_cb(char *arg);
static int disconnect_cb(char *arg);
static int exit_cb(char *arg);
static int help_cb(char *arg);
static int ipc_cb(char *arg);
static int modules_cb(char *arg);
static int source_cb(char *arg);
static int unalias_cb(char *arg);

/* commands */
COMMAND commands[] = {
	{"alias",alias_cb,"Set up an alias for a command."},
	{"connect",connect_cb,"Connect to a Nexxus."},
	{"disconnect",disconnect_cb,"Disconnect from a Nexxus."},
	{"exit",exit_cb,"Exit vash."},
	{"help",help_cb,"Display help for a command."},
	{"ipc",ipc_cb,"Send an IPC message."},
	{"nexxus_modules",modules_cb,"Query a Nexxus for a module listing."},
	{"source",source_cb,"Source the contents of a file."},
	{"unalias",unalias_cb,"Remove a command alias."},
	{(char *)NULL,(Function *)NULL,(char *)NULL}
};

#endif
